package outwatchapp.util.reactive

import scala.concurrent.Future

import io.circe.Decoder
import io.circe.Encoder
import io.circe.Printer
import io.circe.generic.JsonCodec
import io.circe.parser._
import io.circe.syntax._
import monix.bio.Task
import monix.execution.Ack
import monix.execution.Cancelable
import monix.execution.CancelablePromise
import monix.execution.cancelables.SingleAssignCancelable
import monix.reactive.Observable
import monix.reactive.Observer
import monix.reactive.OverflowStrategy
import org.scalajs.dom.raw.Event
import org.scalajs.dom.raw.MessageEvent
import org.scalajs.dom.{raw => sjsdr}
import outwatchapp.util.reactive.MonixProSubject
import monix.reactive.observers.Subscriber
import monix.execution.Scheduler
import monix.reactive.observers.BufferedSubscriber
import monix.execution.cancelables.CompositeCancelable
import scala.concurrent.duration._

class WebSocketImpl[T: Encoder: Decoder](
    ws: sjsdr.WebSocket,
    overflowStrategy: OverflowStrategy.Synchronous[T]
) {
  val printer = Printer.noSpaces

  val source: Task[Observable[T]] =
    Task.deferAction(implicit s =>
      for {
        c <- Task(CompositeCancelable())
        obs <- Task(
          Observable
            .create[T](overflowStrategy) { sub =>
              // val c = SingleAssignCancelable()
              ws.onmessage = (e: MessageEvent) =>
                e.data match {
                  case s: String =>
                    decode[T](s)
                      .map { res =>
                        if (sub.onNext(res) == Ack.Stop) c.cancel()
                        res
                      }
                      .left
                      .foreach(err =>
                        sub.onError(
                          new Exception(s"Failed to decode $s. Error was $err")
                        )
                      )
                  case other =>
                    sub.onError(new Exception(s"Received wrong type: $other"))
                }
              ws.onclose = a => {
                println(
                  s"Websocket closing -  ${a.code} ${a.wasClean} ${a.reason}"
                )
                sub.onComplete()
              }
              ws.onerror = (e: Event) => {
                println("Error in websocket")
                sub.onError(new Exception(s"Error in WebSocket: $e"))
              }
              c += Cancelable { () =>
                println("Closing websocket")
                ws.close()
              }
            }
            .publish
            .refCount
        )
        // empty subscription because otherwise ouwtatch kills the observable when
        // the dom containing this observable is unmounted
        _ <- Task(c += obs.subscribe(Observer.empty))
      } yield obs
    )

  val sink: Task[Subscriber[T]] =
    Task.deferAction(implicit s =>
      Task(
        BufferedSubscriber(
          new Subscriber[T] {
            override implicit def scheduler: Scheduler = s

            override def onNext(elem: T): Future[Ack] = {
              val msg = printer.print(elem.asJson)
              if (ws.readyState == 2 || ws.readyState == 3)
                // Future.failed(new Exception("Websocket closed"))
                Future.successful(Ack.Stop)
              else {
                ws.send(msg)
                Future.successful(Ack.Continue)
              }
            }
            override def onError(ex: Throwable): Unit = println(ex)
            override def onComplete(): Unit = ()
          },
          OverflowStrategy.BackPressure(50)
        )
      )
    )

}
object WebSocket {
  sealed trait Error
  final case object Error extends Error

  def apply[T <: Product: Encoder: Decoder](
      path: String,
      overflowStrategy: OverflowStrategy.Synchronous[T] =
        OverflowStrategy.DropOld(50)
  ): Task[WebSocket[T]] = for {
    websocket <- Task(new sjsdr.WebSocket(path))
    p <- Task(CancelablePromise[Unit]())
    // wait for websocket to open
    _ <- Task(websocket.onopen = _ => p.success(()))
    _ <- Task.deferFuture(p.future).timeout(10.seconds)
    _ <- Task(websocket.onopen = _ => ())
    impl = new WebSocketImpl[T](websocket, overflowStrategy)
    source <- impl.source
    sink <- impl.sink
  } yield MonixProSubject.from(sink, source)
}

@JsonCodec
case class WebsocketData(data: Long)
